/*
 * Copyright (c) 2019, Cisco Systems, Inc.
 *
 * Licensed under the Apache License 2.0 (the "License").  You may not use
 * this file except in compliance with the License.  You can obtain a copy
 * in the file LICENSE in the source distribution or at
 * https://github.com/cisco/libacvp/LICENSE
 */


#ifdef OPENSSL_KDF_SUPPORT
#include <openssl/evp.h>
#include <openssl/bn.h>
#include <openssl/kdf.h>
#include <openssl/tls1.h>
#include <openssl/ssl3.h>
#include "app_lcl.h"
# include "app_fips_lcl.h"

#define TLS_MD_MASTER_SECRET_CONST              "master secret"
#define TLS_MD_MASTER_SECRET_CONST_SIZE         13
#define TLS_MD_KEY_EXPANSION_CONST              "key expansion"
#define TLS_MD_KEY_EXPANSION_CONST_SIZE         13

static int tls1_PRF(const EVP_MD *md,
                    const void *seed1, size_t seed1_len,
                    const void *seed2, size_t seed2_len,
                    const void *seed3, size_t seed3_len,
                    const void *seed4, size_t seed4_len,
                    const void *seed5, size_t seed5_len,
                    const unsigned char *sec, size_t slen,
                    unsigned char *out, size_t olen)
{
    EVP_PKEY_CTX *pctx = NULL;
    int ret = 0;

    if (md == NULL) {
        return 0;
    }
    pctx = EVP_PKEY_CTX_new_id(EVP_PKEY_TLS1_PRF, NULL);
    if (pctx == NULL || EVP_PKEY_derive_init(pctx) <= 0
        || EVP_PKEY_CTX_set_tls1_prf_md(pctx, md) <= 0
        || EVP_PKEY_CTX_set1_tls1_prf_secret(pctx, sec, (int)slen) <= 0
        || EVP_PKEY_CTX_add1_tls1_prf_seed(pctx, seed1, (int)seed1_len) <= 0
        || EVP_PKEY_CTX_add1_tls1_prf_seed(pctx, seed2, (int)seed2_len) <= 0
        || EVP_PKEY_CTX_add1_tls1_prf_seed(pctx, seed3, (int)seed3_len) <= 0
        || EVP_PKEY_CTX_add1_tls1_prf_seed(pctx, seed4, (int)seed4_len) <= 0
        || EVP_PKEY_CTX_add1_tls1_prf_seed(pctx, seed5, (int)seed5_len) <= 0
        || EVP_PKEY_derive(pctx, out, &olen) <= 0) {
        goto err;
    }

    ret = 1;

 err:
    EVP_PKEY_CTX_free(pctx);
    return ret;
}

int app_kdf135_srtp_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int app_kdf135_ikev2_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int app_kdf135_ikev1_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int app_kdf135_x963_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int app_kdf108_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int app_kdf135_tls_handler(ACVP_TEST_CASE *test_case) {
    int rc = 1;
    ACVP_KDF135_TLS_TC *tc = NULL;
    const EVP_MD *md = NULL;

    if (!test_case) {
        return rc;
    }

    tc = test_case->tc.kdf135_tls;
    if (!tc) return rc;

    switch (tc->md) {
    case ACVP_SHA1:
        md = EVP_sha1();
        break;
    case ACVP_SHA224:
        md = EVP_sha224();
        break;
    case ACVP_SHA256:
        md = EVP_sha256();
        break;
    case ACVP_SHA384:
        md = EVP_sha384();
        break;
    case ACVP_SHA512:
        md = EVP_sha512();
        break;
    case ACVP_SHA512_224:
        md = EVP_sha512_224();
        break;
    case ACVP_SHA512_256:
        md = EVP_sha512_256();
        break;
    case ACVP_SHA3_224:
        md = EVP_sha3_224();
        break;
    case ACVP_SHA3_256:
        md = EVP_sha3_256();
        break;
    case ACVP_SHA3_384:
        md = EVP_sha3_384();
        break;
    case ACVP_SHA3_512:
        md = EVP_sha3_512();
        break;
    default:
        printf("Error: Unsupported hash algorithm requested by ACVP server\n");
        return ACVP_NO_CAP;
    }

    //
    // Generate master secret
    //

    rc = tls1_PRF(md,
                  TLS_MD_MASTER_SECRET_CONST,
                  TLS_MD_MASTER_SECRET_CONST_SIZE,
                  tc->ch_rnd, tc->ch_rnd_len,
                  NULL, 0,
                  tc->sh_rnd, tc->sh_rnd_len,
                  NULL, 0,
                  tc->pm_secret, tc->pm_len,
                  tc->msecret1, SSL3_MASTER_SECRET_SIZE);
    if (rc <= 0) {
        printf("Unable to generate master secret\n");
        goto err;
    }

    //
    // Generate key block
    //

    rc = tls1_PRF(md,
                 TLS_MD_KEY_EXPANSION_CONST,
                 TLS_MD_KEY_EXPANSION_CONST_SIZE,
                 tc->s_rnd, tc->s_rnd_len,
                 NULL, 0,
                 tc->c_rnd, tc->c_rnd_len,
                 NULL, 0,
                 tc->msecret1, SSL3_MASTER_SECRET_SIZE,
                 tc->kblock1, tc->kb_len);
    if (rc <= 0) {
        printf("Unable to generate key block\n");
        goto err;
    }

    rc = 0;

err:
    return rc;
}

int app_kdf135_snmp_handler(ACVP_TEST_CASE *test_case) {
    if (!test_case) {
        return -1;
    }
    return 1;
}

int doSSHKDF(const EVP_MD *md,
             unsigned char *input_key,
             size_t input_key_len,
             unsigned char *xcghash,
             size_t xcghash_len,
             unsigned char *session_id,
             size_t session_id_len,
             int action,
             unsigned char *out,
             size_t out_len) {
    int ret = 0;
    EVP_KDF_CTX *kctx;

    if ((kctx = EVP_KDF_CTX_new_id(EVP_KDF_SSHKDF)) == NULL) {
        printf("EVP_KDF_CTX_new_id failed \n");
        goto err;
    }
    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_MD, md) <= 0) {
        printf("EVP_KDF_CTRL_SET_MD failed \n");
        goto err;
    }
    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_KEY, input_key,
                     input_key_len) <= 0) {
        printf("EVP_KDF_CTRL_SET_KEY failed \n");
        goto err;
    }
    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_SSHKDF_XCGHASH, xcghash,
                     xcghash_len) <= 0) {
        printf("EVP_KDF_CTRL_SET_SSHKDF_XCGHASH failed \n");
        goto err;
    }
    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_SSHKDF_SESSION_ID, session_id,
                     session_id_len) <= 0) {
        printf("EVP_KDF_CTRL_SET_SSHKDF_SESSION_ID failed \n");
        goto err;
    }
    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_SSHKDF_TYPE, action) <= 0) {
        printf("EVP_KDF_CTRL_SET_SSHKDF_TYPE failed \n");
        goto err;
    }
    if (EVP_KDF_derive(kctx, out, out_len) <= 0) {
        printf("EVP_KDF_derive failed\n");
        goto err;
    }

    ret = 1;

err:
    EVP_KDF_CTX_free(kctx);

    return ret;
}

int app_kdf135_ssh_handler(ACVP_TEST_CASE *test_case) {
    int rc = 1;
    ACVP_KDF135_SSH_TC *tc = NULL;
    const EVP_MD *md = NULL;

    if (!test_case) {
        return 1;
    }

    tc = test_case->tc.kdf135_ssh;
    if (!tc) return rc;

    switch (tc->sha_type) {
    case ACVP_SHA1:
        md = EVP_sha1();
        break;
    case ACVP_SHA224:
        md = EVP_sha224();
        break;
    case ACVP_SHA256:
        md = EVP_sha256();
        break;
    case ACVP_SHA384:
        md = EVP_sha384();
        break;
    case ACVP_SHA512:
        md = EVP_sha512();
        break;
    case ACVP_SHA512_224:
        md = EVP_sha512_224();
        break;
    case ACVP_SHA512_256:
        md = EVP_sha512_256();
        break;
    case ACVP_SHA3_224:
        md = EVP_sha3_224();
        break;
    case ACVP_SHA3_256:
        md = EVP_sha3_256();
        break;
    case ACVP_SHA3_384:
        md = EVP_sha3_384();
        break;
    case ACVP_SHA3_512:
        md = EVP_sha3_512();
        break;
    default:
        printf("Error: Unsupported hash algorithm requested by ACVP server\n");
        return ACVP_NO_CAP;
    }
    
    //
    // Initial IV client
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'A',
                  tc->cs_init_iv,
                  (size_t)tc->iv_len);
    if (rc <= 0) {
        printf("Unable to get initial IV client\n");
        goto err;
    }

    //
    // Encryption key client
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'C',
                  tc->cs_encrypt_key,
                  (size_t)tc->e_key_len);
    if (rc <= 0) {
        printf("Unable to get encryption key client\n");
        goto err;
    }

    //
    // Integrity key client
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'E',
                  tc->cs_integrity_key,
                  (size_t)tc->i_key_len);
    if (rc <= 0) {
        printf("Unable to get integrity key client\n");
        goto err;
    }

    //
    // Initial IV server
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'B',
                  tc->sc_init_iv,
                  (size_t)tc->iv_len);
    if (rc <= 0) {
        printf("Unable to get initial IV server\n");
        goto err;
    }
    
    //
    // Encryption key server
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'D',
                  tc->sc_encrypt_key,
                  (size_t)tc->e_key_len);
    if (rc <= 0) {
        printf("Unable to get encryption key server\n");
        goto err;
    }

    //
    // Integrity key server
    //

    rc = doSSHKDF(md,
                  tc->shared_secret_k,
                  (size_t)tc->shared_secret_len,
                  tc->hash_h,
                  (size_t)tc->hash_len,
                  tc->session_id,
                  (size_t)tc->session_id_len,
                  (int)'F',
                  tc->sc_integrity_key,
                  (size_t)tc->i_key_len);
    if (rc <= 0) {
        printf("Unable to get integrity key server\n");
        goto err;
    }
    
    rc = 0;
err:
    return rc;
}

int app_pbkdf_handler(ACVP_TEST_CASE *test_case) {
    int rc = 1;
    ACVP_PBKDF_TC *tc = NULL;
    EVP_KDF_CTX *kctx = NULL;
    const EVP_MD *md = NULL;

    if (!test_case) {
        return 1;
    }

    tc = test_case->tc.pbkdf;
    if (!tc) return rc;

    switch (tc->hmac_type) {
    case ACVP_HMAC_ALG_SHA1:
        md = EVP_sha1();
        break;
    case ACVP_HMAC_ALG_SHA224:
        md = EVP_sha224();
        break;
    case ACVP_HMAC_ALG_SHA256:
        md = EVP_sha256();
        break;
    case ACVP_HMAC_ALG_SHA384:
        md = EVP_sha384();
        break;
    case ACVP_HMAC_ALG_SHA512:
        md = EVP_sha512();
        break;
    case ACVP_HMAC_ALG_SHA512_224:
        md = EVP_sha512_224();
        break;
    case ACVP_HMAC_ALG_SHA512_256:
        md = EVP_sha512_256();
        break;
    case ACVP_HMAC_ALG_SHA3_224:
        md = EVP_sha3_224();
        break;
    case ACVP_HMAC_ALG_SHA3_256:
        md = EVP_sha3_256();
        break;
    case ACVP_HMAC_ALG_SHA3_384:
        md = EVP_sha3_384();
        break;
    case ACVP_HMAC_ALG_SHA3_512:
        md = EVP_sha3_512();
        break;
    default:
        printf("Error: Unsupported hash algorithm requested by ACVP server\n");
        return ACVP_NO_CAP;
    }

    if ((kctx = EVP_KDF_CTX_new_id(EVP_KDF_PBKDF2)) == NULL) {
        printf("EVP_KDF_CTX_new_id error\n");
        goto err;
    }

    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_PASS, tc->password, (size_t)tc->pw_len) <=0) {
        printf("EVP_KDF_CTRL_SET_PASS error\n");
        goto err;
    }

    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_SALT, tc->salt, (size_t)tc->salt_len) <= 0) {
        printf("EVP_KDF_CTRL_SET_SALT error\n");
        goto err;
    }

    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_ITER, tc->iterationCount) <= 0) {
        printf("EVP_KDF_CTRL_SET_ITER error\n");
        goto err;
    }

    if (EVP_KDF_ctrl(kctx, EVP_KDF_CTRL_SET_MD, md) <= 0) {
        printf("EVP_KDF_CTRL_SET_MD error\n");
        goto err;
    }

    if (EVP_KDF_derive(kctx, tc->key, tc->key_len) <= 0) {
        printf("EVP_KDF_derive error\n");
        goto err;
    }

    rc = 0;
err:

    if (kctx != NULL) {
        EVP_KDF_CTX_free(kctx);
    }

    return rc;
}

#endif // OPENSSL_KDF_SUPPORT
